/*
 * Copyright (C) 2014, Stephan Mueller <smueller@chronox.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU General Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, ALL OF
 * WHICH ARE HEREBY DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF NOT ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "cryptoperf.h"
#include <linux/debugfs.h>
#include <linux/list.h>

/*
 * This data structure holds the dentry's of the debugfs files establishing
 * the interface to user space.
 */
struct cp_debugfs {
	struct dentry *cp_debugfs_root; /* root dentry */
	struct dentry *cp_debugfs_unused; /* .../unused */
	struct dentry *cp_debugfs_used; /* .../used */
	struct dentry *cp_debugfs_exec; /* .../exec */
};

static struct cp_debugfs cp_debugfs;
/* list of all unused test cases */
static LIST_HEAD(cp_testunused);
/* list of test cases to be executed */
static LIST_HEAD(cp_testexec);
DECLARE_RWSEM(cp_sem);

#define MAXNAME 50

static int cp_debugfs_init_root(void)
{
	cp_debugfs.cp_debugfs_root =
	debugfs_create_dir(DRIVER_NAME, NULL);
	if (IS_ERR(cp_debugfs.cp_debugfs_root)) {
		cp_debugfs.cp_debugfs_root = NULL;
		return -1;
	}
	return 0;
}

int cp_register_test(struct cp_test *test)
{
	struct cp_test *curr_test = NULL;

	down_write(&cp_sem);
	list_for_each_entry(curr_test, &cp_testexec, list) {
		if (!strncmp(curr_test->testname, test->testname,
			     strlen(curr_test->testname))) {
			pr_warn(DRIVER_NAME": test case for cipher %s already "
				"registered -- discarding %s\n",
				curr_test->driver_name, test->testname);
			up_write(&cp_sem);
			return -EINVAL;
		}
	}
	INIT_LIST_HEAD(&test->list);
	list_add(&test->list, &cp_testexec);
	up_write(&cp_sem);

	return 0;
}

static ssize_t cp_print_for_each(struct file *file, char __user *buf,
				 size_t nbytes, loff_t *ppos,
				 struct list_head *tests,
				 void (*func)(struct cp_test *test, char **out))
{
	struct cp_test *curr_test = NULL;
	loff_t pos = *ppos;
	ssize_t total = 0;

	if (pos < 0)
		return -EINVAL;
	if (!nbytes)
		return 0;

	down_read(&cp_sem);
	list_for_each_entry(curr_test, tests, list) {
		char *out = NULL;
		size_t copyout = 0;
		size_t len = 0;

		func(curr_test, &out);
		if (!out)
			continue;
		len = strlen(out);
		copyout = min_t(size_t, nbytes, len);
		if (pos > total) {
			total += copyout;
			kfree(out);
			continue;
		}
		if (copy_to_user(buf+pos+total, out, copyout)) {
			total = -EFAULT;
			kfree(out);
			break;
		}
		kfree(out);
		nbytes -= copyout;
		total += copyout;
		*ppos = pos + total;
		if (0 == nbytes)
			break;
	}
	up_read(&cp_sem);

	return (total - pos);
}

static ssize_t cp_reassign_test(struct file *file, const char __user *buf,
				size_t nbytes, loff_t *ppos,
				struct list_head *from, struct list_head *to)
{
	char name[MAXNAME];
	ssize_t ret = 0;
	struct cp_test *curr_from = NULL;

	if (nbytes > MAXNAME-1)
		return -E2BIG;
	/* clear full data structure */
	memset(name, 0, MAXNAME);
	if (nbytes == 0)
		return nbytes;
	ret = simple_write_to_buffer(name, MAXNAME-1, ppos, buf, nbytes);
	if (0 >= ret)
		return ret;

	while (1) {
		int found = 0;
		down_write(&cp_sem);
		list_for_each_entry(curr_from, from, list) {
			/* match test name */
			if (!strncmp(curr_from->testname,
				     name, strlen(curr_from->testname)) ||
			    /* match test type */
			    !strncmp(curr_from->type,
				     name, strlen(curr_from->type)) ||
			    /* match all if 'all' is provided */
			    (!strncmp(name, "all", 3))) {
				dbg("reassign %s %s\n", curr_from->testname,
				    curr_from->type);
				list_del(&curr_from->list);
				list_add(&curr_from->list, to);
				found = 1;
				break;
			}
		}
		up_write(&cp_sem);
		if (!found)
			break;
	}
	return ret;
}

static void cp_unused_read_fn(struct cp_test *test, char **out)
{
	*out = kasprintf(GFP_KERNEL, "%-50s | %40s | %s\n", test->testname,
			 test->driver_name, test->type);
}

static ssize_t cp_unused_read(struct file *file, char __user *buf,
			      size_t nbytes, loff_t *ppos)
{
	return cp_print_for_each(file, buf, nbytes, ppos, &cp_testunused,
				 cp_unused_read_fn);
}


static ssize_t cp_unused_write(struct file *file, const char __user *buf,
			       size_t nbytes, loff_t *ppos)
{
	return cp_reassign_test(file, buf, nbytes, ppos, &cp_testexec,
				&cp_testunused);
}

static struct file_operations cp_unused_fops = {
	.owner = THIS_MODULE,
	.read = cp_unused_read,
	.write = cp_unused_write,
};

static int cp_debugfs_init_unused(void)
{
	cp_debugfs.cp_debugfs_unused =
	debugfs_create_file("unused", S_IRUGO|S_IWUSR,
			    cp_debugfs.cp_debugfs_root,
		     NULL, &cp_unused_fops);
	if (IS_ERR(cp_debugfs.cp_debugfs_unused)) {
		cp_debugfs.cp_debugfs_unused = NULL;
		return -1;
	}
	return 0;
}

static ssize_t cp_used_read(struct file *file, char __user *buf,
			    size_t nbytes, loff_t *ppos)
{
	return cp_print_for_each(file, buf, nbytes, ppos, &cp_testexec,
				 cp_unused_read_fn);
}

static ssize_t cp_used_write(struct file *file, const char __user *buf,
			     size_t nbytes, loff_t *ppos)
{
	return cp_reassign_test(file, buf, nbytes, ppos, &cp_testunused,
				&cp_testexec);
}

static struct file_operations cp_used_fops = {
	.owner = THIS_MODULE,
	.read = cp_used_read,
	.write = cp_used_write,
};

static int cp_debugfs_init_used(void)
{
	cp_debugfs.cp_debugfs_used =
	debugfs_create_file("used", S_IRUGO|S_IWUSR,
			    cp_debugfs.cp_debugfs_root,
		     NULL, &cp_used_fops);
	if (IS_ERR(cp_debugfs.cp_debugfs_used)) {
		cp_debugfs.cp_debugfs_used = NULL;
		return -1;
	}
	return 0;
}

static void cp_exec_read_fn(struct cp_test *test, char **out)
{
	*out = cp_print_status(test);
}


static ssize_t cp_exec_read(struct file *file, char __user *buf,
			    size_t nbytes, loff_t *ppos)
{
	return cp_print_for_each(file, buf, nbytes, ppos, &cp_testexec,
				 cp_exec_read_fn);
}

static ssize_t cp_exec_write(struct file *file, const char __user *buf,
			     size_t nbytes, loff_t *ppos)
{
	char data[MAXNAME];
	ssize_t ret = 0;
	int rc = 0;
	unsigned int exectime;
	struct cp_test *curr_test = NULL;

	if (nbytes > MAXNAME-1)
		return -E2BIG;
	/* clear full data structure */
	memset(data, 0, MAXNAME);
	if (nbytes == 0)
		return nbytes;
	ret = simple_write_to_buffer(data, MAXNAME-1, ppos, buf, nbytes);
	if (0 >= ret)
		return ret;

	rc = kstrtouint(data, 10, &exectime);
	if (rc)
		return rc;

	down_read(&cp_sem);
	list_for_each_entry(curr_test, &cp_testexec, list)
		cp_exec_test(curr_test, exectime);
	up_read(&cp_sem);

	return ret;
}

static struct file_operations cp_exec_fops = {
	.owner = THIS_MODULE,
	.read = cp_exec_read,
	.write = cp_exec_write,
};

static int cp_debugfs_init_exec(void)
{
	cp_debugfs.cp_debugfs_exec =
	debugfs_create_file("exec", S_IRUGO|S_IWUSR,
			    cp_debugfs.cp_debugfs_root,
		     NULL, &cp_exec_fops);
	if (IS_ERR(cp_debugfs.cp_debugfs_exec)) {
		cp_debugfs.cp_debugfs_exec = NULL;
		return -1;
	}
	return 0;
}

int cp_debugfs_init(void)
{
	int ret = -ENOMEM;

	if(cp_debugfs_init_root() < 0)
		return ret;
	if(cp_debugfs_init_unused() < 0)
		goto outfs;
	if(cp_debugfs_init_used() < 0)
		goto outfs;
	if(cp_debugfs_init_exec() < 0)
		goto outfs;
	return 0;
outfs:
	debugfs_remove_recursive(cp_debugfs.cp_debugfs_root);
	return ret;
}

void cp_debugfs_exit(void)
{
	debugfs_remove_recursive(cp_debugfs.cp_debugfs_root);
}
